#if !defined(AUX_H_INCLUDE_GUARD)

#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdalign.h>


typedef int8_t               S8;
typedef int16_t              S16;
typedef int32_t              S32;
typedef int64_t              S64;
typedef int                  Sint;

typedef uint8_t              U8;
typedef uint16_t             U16;
typedef uint32_t             U32;
typedef uint64_t             U64;
typedef unsigned             Uint;

typedef size_t               Size;

typedef uintptr_t            Uptr;
typedef intptr_t             Sptr;
typedef ptrdiff_t            Dptr;

typedef float                F32;
typedef double               F64;

typedef U8                   B8;
typedef U16                  B16;
typedef U32                  B32;
typedef U64                  B64;
# define true                1U
# define false               0U

typedef unsigned char        Byte;
typedef char                 Char;


# define elemin(array) (sizeof(array)/sizeof((array)[0]))

# if !defined(max)
#  define max(x, y) ((x) >= (y) ? (x) : (y))
# endif

# if !defined(min)
# define min(x, y) ((x) <= (y) ? (x) : (y))
# endif

# define KiB(x) (   (x) * 1024ULL)
# define MiB(x) (KiB(x) * 1024ULL)
# define GiB(x) (MiB(x) * 1024ULL)
# define TiB(x) (GiB(x) * 1024ULL)

# define THOUSAND 1000L
# define MILLION  1000000L
# define BILLION  1000000000L

# define unused_variable(var) (void)var

# define global_variable   static
# define persistent_value  static

# define internal_function static
# define header_function   static inline

#define MEM_MAX_ALIGN_MINUS_ONE (alignof(max_align_t) - 1u)
#define memAlignUp(p) (((p) + MEM_MAX_ALIGN_MINUS_ONE) & (~ MEM_MAX_ALIGN_MINUS_ONE))
#define memAlignDown(p) (mem_ALIGN_UP((p) - MEM_MAX_ALIGN_MINUS_ONE))

typedef enum MemAllocMode {
    MemAllocMode_NONE,
    MemAllocMode_CREATE,
    MemAllocMode_ALLOC,
    MemAllocMode_REALLOC,
    MemAllocMode_DEALLOC,
    MemAllocMode_DEALLOC_ALL,
} MemAllocMode;

# define MEM_ALLOCATOR(allocator)               \
    void* allocator(MemAllocMode mode,          \
                    Size size, void* old_ptr,   \
                    void *data)
typedef MEM_ALLOCATOR(MemAllocator);

struct MemCRT_Header {
    Size size;
};

# define memCRTAlloc(size)        memCRT(MemAllocMode_ALLOC,   size, NULL, NULL)
# define memCRTRealloc(ptr, size) memCRT(MemAllocMode_REALLOC, size, ptr,  NULL)
# define memCRTDealloc(ptr)       memCRT(MemAllocMode_DEALLOC, 0,    ptr,  NULL)

header_function
MEM_ALLOCATOR(memCRT)
{
    unused_variable(data);
    switch (mode) {
        case MemAllocMode_CREATE: {
            // NOTE(naman): Not needed for now
        } break;

        case MemAllocMode_ALLOC: {
            Size memory_size = memAlignUp(size);
            Size header_size = memAlignUp(sizeof(struct MemCRT_Header));
            Size total_size = memory_size + header_size;

            Byte *mem = malloc(total_size);

            struct MemCRT_Header *header = (struct MemCRT_Header *)mem;
            header->size = memory_size;

            Byte *result = mem + header_size;

            memset(result, 0, memory_size);

            return result;
        } break;

        case MemAllocMode_REALLOC: {
            Size memory_size = memAlignUp(size);
            Size header_size = memAlignUp(sizeof(struct MemCRT_Header));
            Size total_size = memory_size + header_size;

            Byte *mem = malloc(total_size);

            struct MemCRT_Header *header = (struct MemCRT_Header *)mem;
            header->size = memory_size;

            Byte *result = mem + header_size;

            if (old_ptr != NULL) {
                Byte *previous_mem = (Byte*)old_ptr - header_size;
                struct MemCRT_Header *previous_header = (struct MemCRT_Header *)previous_mem;
                Size previous_size = previous_header->size;

                memcpy(result, old_ptr, previous_size);
                memset(result + previous_size, 0, memory_size - previous_size);

                memCRTDealloc(old_ptr);
            } else {
                memset(result, 0, memory_size);
            }

            return result;
        } break;

        case MemAllocMode_DEALLOC: {
            if (old_ptr == NULL) {
                return NULL;
            }

            Size header_size = memAlignUp(sizeof(struct MemCRT_Header));
            Byte *mem = (Byte*)old_ptr - header_size;
            free(mem);
        } break;

        case MemAllocMode_DEALLOC_ALL: {
            // TODO(naman): Maybe we should use a off-the-shelf malloc that allows this?
        } break;

        case MemAllocMode_NONE: {
        } break;
    }

    return NULL;
}

/* ==============
 * Strechy Buffer
 */

/* API ----------------------------------------
 * Size  sbufAdd       (T *ptr, T elem)
 * void  sbufDelete    (T *ptr)
 * T*    sbufEnd       (T *ptr)
 *
 * Size  sbufSizeof    (T *ptr)
 * Size  sbufElemin    (T *ptr)
 * Size  sbufMaxSizeof (T *ptr)
 * Size  sbufMaxElemin (T *ptr)
 */

typedef struct Sbuf_Header {
    Size cap; // NOTE(naman): Maximum number of elements that can be stored
    Size len; // NOTE(naman): Count of elements actually stored
    Byte buf[];
} Sbuf_Header;

# define sbuf_GetHeader(sb) ((Sbuf_Header*)(void*)((Byte*)(sb) - offsetof(Sbuf_Header, buf)))

# define sbuf_Len(sb)         ((sb) ? sbuf_GetHeader(sb)->len : 0U)
# define sbuf_Cap(sb)         ((sb) ? sbuf_GetHeader(sb)->cap : 0U)

# define sbufAdd(sb, ...)     ((sb) = sbuf_Grow((sb), sizeof(*(sb))),   \
                               (sb)[sbuf_Len(sb)] = (__VA_ARGS__),      \
                               ((sbuf_GetHeader(sb))->len)++)
# define sbufDelete(sb)       ((sb) ?                                   \
                               (memCRTDealloc(sbuf_GetHeader(sb)), (sb) = NULL) : \
                               0)
# define sbufClear(sb)        ((sb) ?                                   \
                               (memset((sb), 0, sbufSizeof(sb)),        \
                                sbuf_GetHeader(sb)->len = 0) :          \
                               0)
# define sbufResize(sb, n)    (((n) > sbufMaxElemin(sb)) ?              \
                               ((sb) = sbuf_Resize(sb, n, sizeof(*(sb)))) : \
                               0)

# define sbufSizeof(sb)       (sbuf_Len(sb) * sizeof(*(sb)))
# define sbufElemin(sb)       (sbuf_Len(sb))
# define sbufMaxSizeof(sb)    (sbuf_Cap(sb) * sizeof(*(sb)))
# define sbufMaxElemin(sb)    (sbuf_Cap(sb))
# define sbufEnd(sb)          ((sb) + sbuf_Len(sb))

#define sbufPrint(sb, ...) ((sb) = sbuf_Print((sb), __VA_ARGS__))

#define sbufUnsortedDelete(sb, i, z) (((sb)[(i)] = (sb)[sbuf_Len(sb) - 1]), \
                                      ((sb)[sbuf_Len(sb) - 1] = (z)),   \
                                      ((sbuf_GetHeader(sb)->len)--))

# if defined(COMPILER_CLANG)
#  pragma clang diagnostic push
#   pragma clang diagnostic ignored "-Wcast-align"
# endif

header_function
void* sbuf_Grow (void *buf, Size elem_size)
{
    if ((sbuf_Len(buf) + 1) <= sbuf_Cap(buf)) {
        return buf;
    } else {
        Size new_cap = max(2 * sbuf_Cap(buf), 4);

        Size new_size = (new_cap * elem_size) + sizeof(Sbuf_Header);

        Sbuf_Header *new_header = NULL;

        if (buf != NULL) {
            new_header = memCRTRealloc(sbuf_GetHeader(buf), new_size);
        } else {
            new_header = memCRTAlloc(new_size);
        }

        new_header->cap = new_cap;

        return new_header->buf;
    }
}

header_function
void* sbuf_Resize (void *buf, Size elem_count, Size elem_size)
{
    Size new_cap = elem_count;

    Size new_size = (new_cap * elem_size) + sizeof(Sbuf_Header);

    Sbuf_Header *new_header = NULL;

    if (buf != NULL) {
        new_header = memCRTRealloc(sbuf_GetHeader(buf), new_size);
    } else {
        new_header = memCRTAlloc(new_size);
    }

    new_header->cap = new_cap;

    return new_header->buf;
}

header_function
Char* sbuf_Print(Char *buf, const Char *fmt, ...)
{

    va_list args;
    va_start(args, fmt);

    Size cap = sbufMaxElemin(buf) - sbufElemin(buf);
    Size n = 1 + vsnprintf(sbufEnd(buf), cap, fmt, args);
    va_end(args);

    if (n > cap) {
        sbufResize(buf, n + sbufElemin(buf));
        va_start(args, fmt);
        size_t new_cap = sbufMaxElemin(buf) - sbufElemin(buf);
        n = 1 + vsnprintf(sbufEnd(buf), new_cap, fmt, args);
        va_end(args);
    }

    sbuf_GetHeader(buf)->len += (n - 1);
    return buf;
}

header_function
U64 hashFNV1a (Char *str)
{
    U64 hash = 0xcbf29ce484222325; // FNV_offset_basis
    for (Size i = 0; str[i] != '\0'; i++) {
        hash = hash ^ str[i];
        hash = hash * 0x100000001b3; // FNV_prime
    }

    return hash;
}

/* __builtin_clzll(x) returns the leading number of 0-bits in x, starting from
 * most significant position.
 *
 * If b is the bit-width of the number,
 *    p is the closest lower power of two and
 *    lz is the number of leading 0-bits; then
 * then a number between 2^p and 2^(p+1) has the form: (b-p-1 0-bits) 1 (p bits)
 *
 * => lz = b-p-1
 * => p = b-(lz+1)
 *
 * Thus, the rounded-down log2 of the number is b-(lz+1).
 */
header_function
U64 u64Log2(U64 x)
{
    U64 result = 64ULL - ((U64)__builtin_clzll(x) + 1ULL);
    return result;
}

/* Linear Congruential Generator
 *
 * If x is the last random number,
 *    m is a number greater than zero that is a power of two,
 *    a is a number between 0 and m,
 * then the next random number is ((x * a) % m).
 *
 * Unfortunately, the lower bits don't have enought randomness in them. The LSB doesn't
 * change at all, the second LSB alternates, the one after that toggles every 2 turns and so
 * on. Therefore, we try to get rid of the LSBs by pulling in some MSBs.
 *
 * We do the multiplcation twice because Chi-Square Test indicated that this method
 * gives better randomness. Don't ask.
 *
 * NOTE(naman): Seed should be an odd number or the randomness might drop drastically.
 */

header_function
U64 u64Rand (U64 seed)
{
    U64 previous = seed;
    if (previous == 0) {
        // This seed has been tested and should be preferred in normal circumstances.
        previous = 2531011ULL;
    }

    U64 a = 214013ULL;

    __uint128_t product = (__uint128_t)previous * (__uint128_t)a;
    U64 upper = product >> 64, lower = (U64)product;
    U64 log_upper = u64Log2(upper);
    U64 shift_amount = 64 - (log_upper + 1);
    upper = (upper << shift_amount) | (lower >> log_upper);

    U64 result = upper * a;

    return result;
}

/* ==========================
 * Pointer-Pointer Hash Map
 */

typedef struct Hashmap {
    struct HashmapKeys {
        U64  hash;
        Uptr key;
    } *keys;

    Uptr *values;

    MemAllocator *allocator;

    Size total_slots, filled_slots;

    U64 a, b, m; /* Hashing constants */

    U64 r; /* Last random number */
} Hashmap;

typedef enum HM_Flag {
    HM_Flag_EMPTY,
    HM_Flag_FILLED,
    HM_Flag_VACATED, // Previously filled
} HM_Flag;


/* API ---------------------------------------------------
 * Hashmap hmCreate  (MemAllocator allocator,
 *                    Size min_slots);
 * void    hmDelete  (Hashmap hm);
 * void*   hmInsert  (Hashmap *hm, void *key, void *value);
 * Uptr    hmInsertI (Hashmap *hm, Uptr key, Uptr value);
 * void*   hmLookup  (Hashmap *hm, void *key);
 * Uptr    hmLookupI (Hashmap *hm, Uptr key);
 * void*   hmRemove  (Hashmap *hm, void *key);
 * Uptr    hmRemoveI (Hashmap *hm, Uptr key);
 */


#define hm_GetFlag(hash) ((hash) >> 62)
#define hm_GetHash(hash) ((hash) & 0x3FFFFFFFFFFFFFFFULL)
#define hm_SetFlag(hash, flag) (hm_GetHash(hash) | (((U64)(flag)) << 62))

/*
 * https://en.wikipedia.org/wiki/Universal_hashing#Avoiding_modular_arithmetic

 * w is number of bits in machine word (64 in our case)
 * s is the number of buckets/bins (slots in the hash table) to which the
 *   universe of hashable objects is to be mapped
 * m is log2(s) (=> m = 2^s) and is equal to the number of bits in the final hash
 * a is a random odd positive integer < 2^w (fitting in w bits)
 * b is a random non-negative integer < 2^(w-m) (fitting in (w-m) bits)
 *
 * r is the last random number generated and is just an implementation detail.
 */

header_function
void hm_UpdateConstants (Hashmap *hm)
{
    hm->m = hm->m + 1;
    hm->total_slots = 1ULL << (hm->m);

    do {
        hm->r = u64Rand(hm->r);
        hm->a = hm->r;
    } while ((hm->a & 0x01) != 0x01); // Make sure that hm.a is odd

    hm->r = u64Rand(hm->r);
    // b should be (64 - m) bits long
    hm->b = hm->r >> hm->m;
}

header_function
U64 hm_Hash (Hashmap *hm, Uptr key)
{
    U64 result =  ((hm->a * key) + hm->b) & (0xFFFFFFFFFFFFFFFFULL >> (64 - hm->m));
    return result;
}

header_function
Hashmap hmCreate (MemAllocator allocator,
                  Size min_slots)
{
    Hashmap hm = {0};
    hm.allocator = allocator;
    hm.m = u64Log2(min_slots); // Log of closest lower power of two

    // This will make m log of closest upper power of two
    hm_UpdateConstants(&hm);

    hm.keys   = allocator(MemAllocMode_ALLOC,
                          (hm.total_slots) * sizeof(*(hm.keys)),
                          NULL, NULL);
    hm.values = allocator(MemAllocMode_ALLOC,
                          (hm.total_slots) * sizeof(*(hm.values)),
                          NULL, NULL);

    // For NULL keys
    hm.keys[0].hash = hm_SetFlag(hm.keys[0].hash, HM_Flag_FILLED);
    hm.filled_slots = 1;

    return hm;
}

header_function
void hmDelete (Hashmap hm)
{
    hm.allocator(MemAllocMode_DEALLOC, 0, hm.keys,   NULL);
    hm.allocator(MemAllocMode_DEALLOC, 0, hm.values, NULL);
}

header_function
Size hm_LinearProbeSearch (Hashmap *hm, Uptr key)
{
    if (key == 0) return 0;

    U64 hash = hm_Hash(hm, key);
    U64 hash_real = hm_GetHash(hash);

    Size index = 0, i = 0;
    B32 found = false;

    for (i = 0; !found && (i < hm->total_slots); ++i) {
        index = (hash_real + i) % (hm->total_slots);

        HM_Flag flag = hm_GetFlag(hm->keys[index].hash);
        if ((flag == HM_Flag_FILLED) &&
            (hm->keys[index].key == key)) {
            found = true;
        } else if (flag == HM_Flag_VACATED) {
            continue;
        } else if (flag == HM_Flag_EMPTY) {
            break;
        }
    }

    if (i == hm->total_slots) {
        index = 0;
    }

    Size result = (found ? index : 0);

    return result;
}

header_function
Uptr hm_LinearProbeInsertion (Hashmap *hm,
                              U64 hash, Uptr key, Uptr value)
{
    Uptr result_value = value;

    for (Size i = 0; i < hm->total_slots; ++i) {
        Size index = (hash + i) % (hm->total_slots);
        HM_Flag flag  = hm_GetFlag(hm->keys[index].hash);

        B32 fill_now = false;
        switch (flag) {
            case HM_Flag_FILLED: {
                if (hm->keys[index].key == key) {
                    result_value = hm->values[index];
                    fill_now = true;
                }
            } break;
            case HM_Flag_VACATED:
            case HM_Flag_EMPTY: {
                fill_now = true;
            } break;
        }

        if (fill_now) {
            hm->keys[index].key = key;
            hm->values[index] = value;
            hm->keys[index].hash = hm_SetFlag(hash, HM_Flag_FILLED);
            break;
        }
    }

    return result_value;
}

header_function
Uptr hmInsertI (Hashmap *hm, Uptr key, Uptr value)
{
    if ((key == 0) || (value == 0)) return 0;

    if ((2U * (hm->filled_slots)) > (hm->total_slots)) {
        Size total_slots = hm->total_slots;
        struct HashmapKeys *keys   = hm->keys;
        Uptr *values = hm->values;

        hm_UpdateConstants(hm);

        hm->keys   = (hm->allocator)(MemAllocMode_ALLOC,
                                     sizeof(*(hm->keys)) * hm->total_slots,
                                     NULL, NULL);
        hm->values = (hm->allocator)(MemAllocMode_ALLOC,
                                     sizeof(*(hm->values)) * hm->total_slots,
                                     NULL, NULL);

        // For NULL keys
        hm->keys[0].hash = hm_SetFlag(hm->keys[0].hash, HM_Flag_FILLED);

        for (Size i = 1; i < total_slots; ++i) {
            U64 hash_i_old = keys[i].hash;
            Uptr key_i     = keys[i].key;
            Uptr value_i   = values[i];
            HM_Flag flag_i   = hm_GetFlag(hash_i_old);
            U64 hash_i_new = hm_Hash(hm, key_i);
            if (flag_i == HM_Flag_FILLED) {
                hm_LinearProbeInsertion(hm, hash_i_new, key_i, value_i);
            }
        }

        (hm->allocator)(MemAllocMode_DEALLOC, 0, keys,   NULL);
        (hm->allocator)(MemAllocMode_DEALLOC, 0, values, NULL);
    }

    U64 hash = hm_Hash(hm, key);

    Uptr result_value = hm_LinearProbeInsertion(hm, hash, key, value);
    hm->filled_slots += 1;

    return result_value;
}

# if defined(COMPILER_CLANG)
#  pragma clang diagnostic push
#   pragma clang diagnostic ignored "-Wbad-function-cast"
# endif

header_function
void* hmInsert (Hashmap *hm, void *key, void *value)
{
    void *result_value = (void*)hmInsertI(hm, (Uptr)key, (Uptr)value);
    return result_value;
}

header_function
Uptr hmLookupI (Hashmap *hm, Uptr key)
{
    Size location = hm_LinearProbeSearch(hm, key);
    Uptr result_value = hm->values[location];

    return result_value;
}

header_function
void* hmLookup (Hashmap *hm, void *key)
{
    void *result_value = (void*)hmLookupI(hm, (Uptr)key);
    return result_value;
}

header_function
Uptr hmRemoveI (Hashmap *hm, Uptr key)
{
    Size location = hm_LinearProbeSearch(hm, key);

    Uptr result_value = 0;
    if (location != 0) {
        U64 hash = hm->keys[location].hash;
        hm->keys[location].hash = hm_SetFlag(hash, HM_Flag_VACATED);
        hm->filled_slots -= 1;
        result_value = hm->values[location];
    }

    return result_value;
}

header_function
void* hmRemove (Hashmap *hm, void *key)
{
    void *result_value = (void*)hmRemoveI(hm, (Uptr)key);
    return result_value;
}

header_function
Size strprefix(Char *str, Char *pre)
{
    Size lenpre = strlen(pre);
    Size lenstr = strlen(str);

    if (lenstr < lenpre) {
        return 0;
    } else {
        if (memcmp(pre, str, lenpre) == 0) {
            return lenpre;
        } else {
            return 0;
        }
    }
}

header_function
B32 strsuffix (Char *str, Char *suf)
{
    Char *string = strrchr(str, suf[0]);
    B32 result = false;

    if(string != NULL) {
        if (strcmp(string, suf) == 0) {
            result = true;
        }
    }

    return result;
}

header_function
U64 nextPowerOf2(U64 v)
{
    v--;
    v |= v >> 1;
    v |= v >> 2;
    v |= v >> 4;
    v |= v >> 8;
    v |= v >> 16;
    v |= v >> 32;
    v++;
    return v;

}


#define AUX_H_INCLUDE_GUARD
#endif
