typedef enum XML_Message_Kind {
    XML_Message_Kind_NONE,
    XML_Message_Kind_GET,
    XML_Message_Kind_PUT,
    XML_Message_Kind_DELETE,
    XML_Message_Kind_RESP_,
    XML_Message_Kind_RESP_GET,
    XML_Message_Kind_RESP_PUT_DELETE,
    XML_Message_Kind_RESP_ERROR,

    XML_Message_Kind_SYSTEM_ERROR,

    XML_Message_Kind_TOTAL
} XML_Message_Kind;

typedef struct XML_Message {
    XML_Message_Kind kind;
    Char *key;
    Char *value;
    Char *error;
} XML_Message;

internal_function
XML_Message xmlParseMessage (Char *xml) {
    Char *stream = xml;
    XML_Message result = {0};
    Char *error_message = "XML Error: Received unparseable message";

#define PARSE_CHAR(c) do { if (stream[0] != c) {goto parse_failed;} else stream++;} while (0)
#define PARSE_SPACE() do { while (isspace(stream[0])) { stream++; }} while (0)

    PARSE_CHAR('<');
    PARSE_CHAR('?');
    PARSE_SPACE();
    PARSE_CHAR('x');
    PARSE_CHAR('m');
    PARSE_CHAR('l');
    PARSE_SPACE();
    PARSE_CHAR('v');
    PARSE_CHAR('e');
    PARSE_CHAR('r');
    PARSE_CHAR('s');
    PARSE_CHAR('i');
    PARSE_CHAR('o');
    PARSE_CHAR('n');
    PARSE_SPACE();
    PARSE_CHAR('=');
    PARSE_SPACE();
    PARSE_CHAR('"');
    PARSE_CHAR('1');
    PARSE_CHAR('.');
    PARSE_CHAR('0');
    PARSE_CHAR('"');
    PARSE_SPACE();
    PARSE_CHAR('e');
    PARSE_CHAR('n');
    PARSE_CHAR('c');
    PARSE_CHAR('o');
    PARSE_CHAR('d');
    PARSE_CHAR('i');
    PARSE_CHAR('n');
    PARSE_CHAR('g');
    PARSE_SPACE();
    PARSE_CHAR('=');
    PARSE_SPACE();
    PARSE_CHAR('"');
    PARSE_CHAR('U');
    PARSE_CHAR('T');
    PARSE_CHAR('F');
    PARSE_CHAR('-');
    PARSE_CHAR('8');
    PARSE_CHAR('"');
    PARSE_SPACE();
    PARSE_CHAR('?');
    PARSE_CHAR('>');

    PARSE_SPACE();

    PARSE_CHAR('<');
    PARSE_SPACE();
    PARSE_CHAR('K');
    PARSE_CHAR('V');
    PARSE_CHAR('M');
    PARSE_CHAR('e');
    PARSE_CHAR('s');
    PARSE_CHAR('s');
    PARSE_CHAR('a');
    PARSE_CHAR('g');
    PARSE_CHAR('e');
    PARSE_SPACE();
    PARSE_CHAR('t');
    PARSE_CHAR('y');
    PARSE_CHAR('p');
    PARSE_CHAR('e');
    PARSE_SPACE();
    PARSE_CHAR('=');
    PARSE_SPACE();
    PARSE_CHAR('"');

    Char kind_str[7] = {0};

    Size kind_str_len = 0;
    while (stream[0] != '"') {
        kind_str[kind_str_len] = stream[0];
        stream++;

        if (kind_str_len >= 6) {
            goto parse_failed;
        }

        kind_str_len++;
    }

    XML_Message_Kind kind = XML_Message_Kind_NONE;

    if (strcmp(kind_str, "getreq") == 0) {
        kind = XML_Message_Kind_GET;
    } else if (strcmp(kind_str, "putreq") == 0) {
        kind = XML_Message_Kind_PUT;
    } else if (strcmp(kind_str, "delreq") == 0) {
        kind = XML_Message_Kind_DELETE;
    } else if (strcmp(kind_str, "resp") == 0) {
        kind = XML_Message_Kind_RESP_;
    } else {
        goto parse_failed;
    }

    PARSE_CHAR('"');
    PARSE_SPACE();
    PARSE_CHAR('>');
    PARSE_SPACE();

    switch (kind) {
        case XML_Message_Kind_GET: {
            result.kind = kind;

            PARSE_CHAR('<');
            PARSE_SPACE();
            PARSE_CHAR('K');
            PARSE_CHAR('e');
            PARSE_CHAR('y');
            PARSE_SPACE();
            PARSE_CHAR('>');

            Size key_len_max = 256;
            result.key = calloc(key_len_max + 1, sizeof(*result.key));
            Size key_len = 0;

            while ((stream[0] != '<') || ((stream[0] == '<') && (stream[-1] == '\\'))) {
                result.key[key_len] = stream[0];
                stream++;

                if (key_len >= key_len_max) {
                    error_message = "Oversized key";
                    goto parse_failed;
                }

                key_len++;
            }

            if (key_len == 0) {
                    error_message = "Empty key";
                    goto parse_failed;
            }

            PARSE_CHAR('<');
            PARSE_SPACE();
            PARSE_CHAR('/');
            PARSE_SPACE();
            PARSE_CHAR('K');
            PARSE_CHAR('e');
            PARSE_CHAR('y');
            PARSE_SPACE();
            PARSE_CHAR('>');
            PARSE_SPACE();
            PARSE_CHAR('<');
            PARSE_SPACE();
            PARSE_CHAR('/');
            PARSE_SPACE();
            PARSE_CHAR('K');
            PARSE_CHAR('V');
            PARSE_CHAR('M');
            PARSE_CHAR('e');
            PARSE_CHAR('s');
            PARSE_CHAR('s');
            PARSE_CHAR('a');
            PARSE_CHAR('g');
            PARSE_CHAR('e');
            PARSE_SPACE();
            PARSE_CHAR('>');
        } break;

        case XML_Message_Kind_PUT: {
            result.kind = kind;

            PARSE_CHAR('<');
            PARSE_SPACE();
            PARSE_CHAR('K');
            PARSE_CHAR('e');
            PARSE_CHAR('y');
            PARSE_SPACE();
            PARSE_CHAR('>');

            Size key_len_max = 256;
            result.key = calloc(key_len_max + 1, sizeof(*result.key));
            Size key_len = 0;

            while ((stream[0] != '<') || ((stream[0] == '<') && (stream[-1] == '\\'))) {
                result.key[key_len] = stream[0];
                stream++;

                if (key_len >= key_len_max) {
                    error_message = "Oversized key";
                    goto parse_failed;
                }

                key_len++;
            }

            PARSE_CHAR('<');
            PARSE_SPACE();
            PARSE_CHAR('/');
            PARSE_SPACE();
            PARSE_CHAR('K');
            PARSE_CHAR('e');
            PARSE_CHAR('y');
            PARSE_SPACE();
            PARSE_CHAR('>');
            PARSE_SPACE();
            PARSE_CHAR('<');
            PARSE_SPACE();
            PARSE_CHAR('V');
            PARSE_CHAR('a');
            PARSE_CHAR('l');
            PARSE_CHAR('u');
            PARSE_CHAR('e');
            PARSE_SPACE();
            PARSE_CHAR('>');

            Size value_len_max = KiB(256);
            result.value = calloc(value_len_max + 1, sizeof(*result.value));
            Size value_len = 0;

            while ((stream[0] != '<') || ((stream[0] == '<') && (stream[-1] == '\\'))) {
                result.value[value_len] = stream[0];
                stream++;

                if (value_len >= value_len_max) {
                    error_message = "Oversized value";
                    goto parse_failed;
                }

                value_len++;
            }

            if (value_len == 0) {
                    error_message = "Empty value";
                    goto parse_failed;
            }

            PARSE_CHAR('<');
            PARSE_SPACE();
            PARSE_CHAR('/');
            PARSE_SPACE();
            PARSE_CHAR('V');
            PARSE_CHAR('a');
            PARSE_CHAR('l');
            PARSE_CHAR('u');
            PARSE_CHAR('e');
            PARSE_SPACE();
            PARSE_CHAR('>');
            PARSE_SPACE();
            PARSE_CHAR('<');
            PARSE_SPACE();
            PARSE_CHAR('/');
            PARSE_SPACE();
            PARSE_CHAR('K');
            PARSE_CHAR('V');
            PARSE_CHAR('M');
            PARSE_CHAR('e');
            PARSE_CHAR('s');
            PARSE_CHAR('s');
            PARSE_CHAR('a');
            PARSE_CHAR('g');
            PARSE_CHAR('e');
            PARSE_SPACE();
            PARSE_CHAR('>');
        } break;

        case XML_Message_Kind_DELETE: {
            result.kind = kind;

            PARSE_CHAR('<');
            PARSE_SPACE();
            PARSE_CHAR('K');
            PARSE_CHAR('e');
            PARSE_CHAR('y');
            PARSE_SPACE();
            PARSE_CHAR('>');

            Size key_len_max = 256;
            result.key = calloc(key_len_max + 1, sizeof(*result.key));
            Size key_len = 0;

            while ((stream[0] != '<') || ((stream[0] == '<') && (stream[-1] == '\\'))) {
                result.key[key_len] = stream[0];
                stream++;

                if (key_len >= key_len_max) {
                    error_message = "Oversized key";
                    goto parse_failed;
                }

                key_len++;
            }

            PARSE_CHAR('<');
            PARSE_SPACE();
            PARSE_CHAR('/');
            PARSE_SPACE();
            PARSE_CHAR('K');
            PARSE_CHAR('e');
            PARSE_CHAR('y');
            PARSE_SPACE();
            PARSE_CHAR('>');
            PARSE_SPACE();
            PARSE_CHAR('<');
            PARSE_SPACE();
            PARSE_CHAR('/');
            PARSE_SPACE();
            PARSE_CHAR('K');
            PARSE_CHAR('V');
            PARSE_CHAR('M');
            PARSE_CHAR('e');
            PARSE_CHAR('s');
            PARSE_CHAR('s');
            PARSE_CHAR('a');
            PARSE_CHAR('g');
            PARSE_CHAR('e');
            PARSE_SPACE();
            PARSE_CHAR('>');
        } break;

        case XML_Message_Kind_RESP_: {
            PARSE_CHAR('<');
            PARSE_SPACE();

            switch (stream[0]) {
                case 'K': {
                    result.kind = XML_Message_Kind_RESP_GET;

                    PARSE_CHAR('K');
                    PARSE_CHAR('e');
                    PARSE_CHAR('y');
                    PARSE_SPACE();
                    PARSE_CHAR('>');

                    Size key_len_max = 256;
                    result.key = calloc(key_len_max + 1, sizeof(*result.key));
                    Size key_len = 0;

                    while ((stream[0] != '<') || ((stream[0] == '<') && (stream[-1] == '\\'))) {
                        result.key[key_len] = stream[0];
                        stream++;

                        if (key_len >= key_len_max) {
                            error_message = "Oversized key";
                            goto parse_failed;
                        }

                        key_len++;
                    }

                    PARSE_CHAR('<');
                    PARSE_SPACE();
                    PARSE_CHAR('/');
                    PARSE_SPACE();
                    PARSE_CHAR('K');
                    PARSE_CHAR('e');
                    PARSE_CHAR('y');
                    PARSE_SPACE();
                    PARSE_CHAR('>');
                    PARSE_SPACE();
                    PARSE_CHAR('<');
                    PARSE_SPACE();
                    PARSE_CHAR('V');
                    PARSE_CHAR('a');
                    PARSE_CHAR('l');
                    PARSE_CHAR('u');
                    PARSE_CHAR('e');
                    PARSE_SPACE();
                    PARSE_CHAR('>');

                    Size value_len_max = KiB(256);
                    result.value = calloc(value_len_max + 1, sizeof(*result.value));
                    Size value_len = 0;

                    while ((stream[0] != '<') || ((stream[0] == '<') && (stream[-1] == '\\'))) {
                        result.value[value_len] = stream[0];
                        stream++;

                        if (value_len >= value_len_max) {
                            error_message = "Oversized value";
                            goto parse_failed;
                        }

                        value_len++;
                    }

                    PARSE_CHAR('<');
                    PARSE_SPACE();
                    PARSE_CHAR('/');
                    PARSE_SPACE();
                    PARSE_CHAR('V');
                    PARSE_CHAR('a');
                    PARSE_CHAR('l');
                    PARSE_CHAR('u');
                    PARSE_CHAR('e');
                    PARSE_SPACE();
                    PARSE_CHAR('>');
                    PARSE_SPACE();
                    PARSE_CHAR('<');
                    PARSE_SPACE();
                    PARSE_CHAR('/');
                    PARSE_SPACE();
                    PARSE_CHAR('K');
                    PARSE_CHAR('V');
                    PARSE_CHAR('M');
                    PARSE_CHAR('e');
                    PARSE_CHAR('s');
                    PARSE_CHAR('s');
                    PARSE_CHAR('a');
                    PARSE_CHAR('g');
                    PARSE_CHAR('e');
                    PARSE_SPACE();
                    PARSE_CHAR('>');
                } break;

                case 'M': {
                    PARSE_CHAR('M');
                    PARSE_CHAR('e');
                    PARSE_CHAR('s');
                    PARSE_CHAR('s');
                    PARSE_CHAR('a');
                    PARSE_CHAR('g');
                    PARSE_CHAR('e');
                    PARSE_SPACE();
                    PARSE_CHAR('>');

                    if ((stream[0] == 'S') &&
                        (stream[1] == 'u') &&
                        (stream[2] == 'c') &&
                        (stream[3] == 'c') &&
                        (stream[4] == 'e') &&
                        (stream[5] == 's') &&
                        (stream[6] == 's')) {
                        result.kind = XML_Message_Kind_RESP_PUT_DELETE;

                        PARSE_CHAR('S');
                        PARSE_CHAR('u');
                        PARSE_CHAR('c');
                        PARSE_CHAR('c');
                        PARSE_CHAR('e');
                        PARSE_CHAR('s');
                        PARSE_CHAR('s');
                        PARSE_CHAR('<');
                        PARSE_SPACE();
                        PARSE_CHAR('/');
                        PARSE_SPACE();
                        PARSE_CHAR('M');
                        PARSE_CHAR('e');
                        PARSE_CHAR('s');
                        PARSE_CHAR('s');
                        PARSE_CHAR('a');
                        PARSE_CHAR('g');
                        PARSE_CHAR('e');
                        PARSE_SPACE();
                        PARSE_CHAR('>');
                        PARSE_SPACE();
                        PARSE_CHAR('<');
                        PARSE_SPACE();
                        PARSE_CHAR('/');
                        PARSE_SPACE();
                        PARSE_CHAR('K');
                        PARSE_CHAR('V');
                        PARSE_CHAR('M');
                        PARSE_CHAR('e');
                        PARSE_CHAR('s');
                        PARSE_CHAR('s');
                        PARSE_CHAR('a');
                        PARSE_CHAR('g');
                        PARSE_CHAR('e');
                        PARSE_SPACE();
                        PARSE_CHAR('>');
                    } else {
                        result.kind = XML_Message_Kind_RESP_ERROR;

                        Size error_len = 0;
                        Char *error_begin = stream;

                        while ((stream[0] != '<') || ((stream[0] == '<') && (stream[-1] == '\\'))) {
                            stream++;
                            error_len++;
                        }

                        result.error = calloc(error_len + 1, sizeof(*result.error));
                        for (Size i = 0; i < error_len; i++) {
                            result.error[i] = error_begin[i];
                        }
                    }
                }  break;

                default: {
                    goto parse_failed;
                }
            }
        } break;

        default: {
            goto parse_failed;
        } break;
    }

#undef PARSE_CHAR
#undef PARSE_SPACE

    return result;

parse_failed:
    free(result.key);
    free(result.value);
    free(result.error);
    return (XML_Message){.kind = XML_Message_Kind_SYSTEM_ERROR, .error = strdup(error_message)};
}

internal_function
Char* xmlCreateMessage (XML_Message_Kind kind,
                        Char *key, Char *value, Char *error) {
    Char *message = NULL;

    switch (kind) {
        case XML_Message_Kind_GET: {
            sbufPrint(message,
                      "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                      "<KVMessage type=\"getreq\">\n"
                      "<Key>");
            sbufPrint(message, "%s", key);
            sbufPrint(message,
                      "</Key>\n"
                      "</KVMessage>\n");
        } break;

        case XML_Message_Kind_PUT: {
            sbufPrint(message,
                      "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                      "<KVMessage type=\"putreq\">\n"
                      "<Key>");
            sbufPrint(message, "%s", key);
            sbufPrint(message,
                      "</Key>\n"
                      "<Value>");
            sbufPrint(message, "%s", value);
            sbufPrint(message,
                      "</Value>\n"
                      "</KVMessage>\n");
        } break;

        case XML_Message_Kind_DELETE: {
            sbufPrint(message,
                      "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                      "<KVMessage type=\"delreq\">\n"
                      "<Key>");
            sbufPrint(message, "%s", key);
            sbufPrint(message,
                      "</Key>\n"
                      "</KVMessage>\n");
        } break;

        case XML_Message_Kind_RESP_GET: {
            sbufPrint(message,
                      "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                      "<KVMessage type=\"resp\">\n"
                      "<Key>");
            sbufPrint(message, "%s", key);
            sbufPrint(message,
                      "</Key>\n"
                      "<Value>");
            sbufPrint(message, "%s", value);
            sbufPrint(message,
                      "</Value>\n"
                      "</KVMessage>\n");
        } break;

        case XML_Message_Kind_RESP_PUT_DELETE: {
            sbufPrint(message,
                      "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                      "<KVMessage type=\"resp\">\n"
                      "<Message>Success</Message>\n"
                      "</KVMessage>\n");
        } break;

        case XML_Message_Kind_RESP_ERROR: {
            sbufPrint(message,
                      "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                      "<KVMessage type=\"resp\">\n"
                      "<Message>");
            sbufPrint(message, "%s", error);
            sbufPrint(message,
                      "</Message>\n"
                      "</KVMessage>\n");
        } break;

        default: {
            message = NULL;
        } break;
    }

    return message;
}
