typedef struct Result {
    struct Result *next;

    union {
        struct {
            Char *value;
            Char *key;
        } get;

        struct {
            Char *error;
        } error;
    };

    enum Result_Kind {
        Result_Kind_NONE,
        Result_Kind_GET,
        Result_Kind_PUT,
        Result_Kind_DEL,
        Result_Kind_ERROR
    } kind;

    Sint fd;
} Result;

global_variable Result *global_result_first, *global_result_last, *global_result_divider;
global_variable pthread_mutex_t global_result_lock = PTHREAD_MUTEX_INITIALIZER;

internal_function
void resultInitBuffer (void)
{
    global_result_first = calloc(1, sizeof(*global_result_first));
    global_result_last = global_result_first;
    global_result_divider = global_result_first;
}

internal_function
void resultEnqueue (Result c)
{
    pthread_mutex_lock(&global_result_lock);

    global_result_last->next = calloc(1, sizeof(*(global_result_last->next)));
    *(global_result_last->next) = c;
    global_result_last = global_result_last->next;

    while (global_result_first != global_result_divider) {
        Result *temp = global_result_first;
        global_result_first = global_result_first->next;
        free(temp);
    }

    pthread_mutex_unlock(&global_result_lock);

    return;
}

internal_function
B32 resultDequeue (Result *c)
{
    if (global_result_divider != global_result_last) {
        *c = *(global_result_divider->next);
        global_result_divider = global_result_divider->next;
        return true;
    }

    return false;
}
